# Main
# ====

# main.py

import os
import time
import random
import pygame
from pygame.locals import *
pygame.mixer.pre_init(44100, -16, 2, 1024)
pygame.mixer.init()
pygame.init()
WIDTH = 1000
HEIGHT = 700

S = pygame.display.set_mode((WIDTH, HEIGHT))
pygame.display.set_caption('Stable Orbit')
FPS = 25
TDFPS = 120
MFPS = 125
Clock = pygame.time.Clock()

def convert(filename):
    filename.replace('/', os.path.sep)
    return filename

def between(points):
    return ((abs(points[0][0] - points[1][0]) ** 2) + (abs(points[0][1] - points[1][1]) ** 2)) ** 0.5

def Stop():
    print('Game Quit')
    pygame.quit()
    raise SystemExit

def PointInsideRect(point, rect):
    if rect[0] < point[0] < (rect[0] + rect[2]) and rect[1] < point[1] < (rect[1] + rect[3]):
        return True
    return False

def CS(bods, num):
    print(str(num) + ' ' + str(bodies[2].speed))

def Middle(surface):
    return round((WIDTH - surface.get_width()) / 2)

FONTFL = pygame.font.Font(convert('data/denmark.ttf'), 90)

DEATHZONE = 2000
PSIZE = 70
pygame.mixer_music.load(convert('data/Phantom From Space.ogg'))
pygame.mixer_music.play(-1)

CLICK = pygame.mixer.Sound('data/Click.ogg')
EXPLOSIONSOUND = pygame.mixer.Sound('data/Explosion.ogg')

LOADING = FONTFL.render('Loading...', 1, (127, 127, 127))
S.blit(LOADING, (Middle(LOADING), 200))
pygame.display.update()


story = '''A supervolcano has erupted. Tens of millions are already
dead. It is only a matter of time before ash blots out the sun
worldwide.

Humanity's only hope is The Escape: an ambitious plan to evade
extinction. The moon is not an option as there is no atmosphere.
Another moon, Moon II, must be created in orbit of the Earth to
live on until the ash clears.

You are an orbital mechanics expert working on The Escape.
Your computer simulations can accurately predict Moon II's
orbit but your time is limited. A collision will kill billions.
A loose orbit will leave them to starve and freeze in the inky
blackness of outer space or burn to death in the fringes of the
sun.

[Press any key to return to the main menu]'''
# Maximum width
###############################################################
instructions = '''Use the mouse to drag the new moon to its desired location.
Only the new moon's location can be adjusted.

The arrow keys can be used to adjust the starting velocity of
the new moon. The current velocity is shown by the red marker
and is equivelant to the distance and direction travelled by
the new moon in one second. This means that the futher away
from the new moon it is, the faster the new moon will move. The
direction of the line shows the direction of travel.

There is a limit to the maximum velocity at which the new moon
can be launched.

Press SPACE or ENTER to run the simulation of the new moon's
orbit. Press SHIFT-ENTER or SHIFT-SPACE to launch for real.
Press ESC during a simulation to abort it. Don't forget that
you're time is limited.


[Press any key to return to the main menu]'''


# Load images
BIGNESS = 6000
BACKGROUND = pygame.surface.Surface((BIGNESS, BIGNESS))
bg = pygame.PixelArray(BACKGROUND)
for i in range(0, round((BIGNESS ** 2) * 0.0002), 1):
    bg[random.randint(0, BIGNESS - 1)][random.randint(0, BIGNESS - 1)] = (255, 255, 255)
del bg

DEIMOS = pygame.image.load(convert('data/Deimos.png'))
PHOBOS = pygame.image.load(convert('data/Phobos.png'))
MARS = pygame.image.load(convert('data/Mars.png'))
MOON = pygame.image.load(convert('data/Moon.png'))
EARTH = pygame.image.load(convert('data/Earth.png'))
VENUS = pygame.image.load(convert('data/Venus.png'))
NEWMOON = pygame.image.load(convert('data/NewMoon.png'))
EXPLOSION = pygame.image.load(convert('data/Explosion.png'))

EarthMass = 5.972 * (10 ** 24)
MoonMass = 7.348 * (10 ** 22)
VenusMass = 4.8676 * (10 ** 24)
MarsMass = 6.417 * (10 ** 23)
PhobosMass = 1.0659 * (10 ** 16)
DeimosMass = 1.4762 * (10 ** 15)

EarthSize = 6371
MoonSize = 1738
VenusSize = 6052
MarsSize = 3390
PhobosSize = 11
DeimosSize = 6

def TextToPicture(text, colour, size, font, btwn):
    FONT = pygame.font.Font(font, size)
    lines = text.split('\n')
    laspic = []
    maxwidth = 0
    for line in lines:
        img = FONT.render(line, 1, colour)
        if img.get_width() > maxwidth:
            maxwidth = img.get_width()

        laspic.append(img)
    pal = pygame.surface.Surface((maxwidth, btwn * len(laspic)))
    for i in range(0, len(laspic), 1):
        pal.blit(laspic[i], (0, i * btwn))
    return pal

class Body():
    def __init__(self, texture, size, mass, position, speed, main=False):
        self.size = size / EarthSize
        self.mass = mass
        self.position = position
        self.speed = speed #Should be velocity
        self.texture = pygame.transform.scale(texture, (round(PSIZE * self.size), round(PSIZE * self.size)))
        self.main = main

    def SetPos(self, newpos):
        self.position = [newpos[0], newpos[1]]

    def SetSpeed(self, newspeed):
        self.speed = [newspeed[0], newspeed[1]]

    def BackUp(self):
        self.bpos = self.position[:]
        self.bspeed = self.speed[:]

    def Restore(self):
        self.position = self.bpos[:]
        self.speed = self.bspeed[:]

    def Clean(self):
        del self.bpos
        del self.bspeed


def Move(bodies):
    for body in bodies:
        for effect in bodies:
            if body != effect:
                try:
                    force = ((body.mass * effect.mass) / ((between((body.position, effect.position))) ** 2)) / (2 * (10 ** 21))
                except ZeroDivisionError:
                    return ['Dead'] + bodies
                accel = force / body.mass
                diff = [body.position[0] - effect.position[0], body.position[1] - effect.position[1]]
                total = abs(diff[0]) + abs(diff[1])
                diff[0] *= accel / total
                diff[1] *= accel / total
                body.speed[0] -= diff[0]
                body.speed[1] -= diff[1]
                if between((body.position, effect.position)) <= (body.size + effect.size) * (PSIZE / 2):
                    location = [(body.position[0] * effect.size + effect.position[0] * body.size) / (effect.size + body.size), (body.position[1] * effect.size + effect.position[1] * body.size) / (effect.size + body.size)]
                    return ['Crash', location] + bodies
                elif not (-DEATHZONE + (WIDTH // 2)) < body.position[0] < DEATHZONE or not (-DEATHZONE + WIDTH / 2) < body.position[1] < DEATHZONE:
                    return ['Out'] + bodies

    for body in bodies:
        body.position[0] += body.speed[0]
        body.position[1] += body.speed[1]
    return bodies


def Level(bodies, real=False):
    timelim = 200
    ftimelim = 6000
    SD = 120
    ext = False
    dead = False
    scan = [0, 0]
    draw = False
    end = -SD
    for body in bodies:
        body.BackUp()
    while not ext:
        for event in pygame.event.get():
            if event.type == KEYDOWN:
                if event.key == K_ESCAPE and not real:
                    CLICK.play()
                    ext = True
            elif event.type == QUIT:
                Stop()

        bodies = Move(bodies)
        if bodies[0] == 'Crash':
            dead = 1
            bodies.remove('Crash')
            loc = bodies.pop(0)

        elif bodies[0] == 'Out':
            dead = 2
            bodies.remove('Out')

        if (timelim > 0 or ftimelim % (round((min(abs(timelim) / 8, 20))) + 1) == 0 or ftimelim - end < SD) and draw:
            S.blit(BACKGROUND, (scan[0] - BIGNESS // 2, scan[1] - BIGNESS // 2))
            for body in bodies:
                if body.main and body.position != [WIDTH // 2, HEIGHT // 2]:
                    dif = [(WIDTH // 2) - body.position[0], (HEIGHT // 2) - body.position[1]]
                    for bod in bodies:
                        bod.position[0] += dif[0]
                        bod.position[1] += dif[1]
                    scan[0] -= dif[0]
                    scan[1] -= dif[1]
                S.blit(body.texture, (round((body.position[0] - (body.size * (PSIZE / 2)))), round(body.position[1] - (body.size * (PSIZE / 2)))))
            if not real:
                S.blit(sim, (10, 10))
                if timelim < 50 and not ftimelim - end < SD:
                    S.blit(hyp, (10, 650))
            elif timelim < 50 and not ftimelim - end < SD:
                S.blit(ffw, (10, 650))
            pygame.display.update()
            Clock.tick(FPS)
        if dead:
            if draw:
                if dead == 1:
                    return [dead, loc, bodies]
                else:

                    return [dead, bodies]
            else:
                end = ftimelim
                for body in bodies:
                    body.Restore()
                draw = True
                dead = False
                scan = [0, 0]
                timelim = 250
                ftimelim = 6000

        timelim -= 1
        ftimelim -= 1
        if ftimelim <= 0:
            if draw:
                return ['Success', bodies]
            else:
                for body in bodies:
                    body.Restore()
                draw = True
                dead = False
                scan = [0, 0]
                timelim = 250
                ftimelim = 6000


levels = [(MoonSize, MoonMass / 10),
          (MoonSize * 1.4, (MoonMass / 10) * (1.4 ** 2)),
          (MoonSize * 1.8, (MoonMass / 10) * (1.8 ** 2)),
          (MoonSize * 2.2, (MoonMass / 10) * (2.2 ** 2))]

lts = [300,
       260,
       240,
       210]

nl = [3, 4, 3]

sposs = [[500, 450],
         [600, 500],
         [250, 250],
         [900, 200]]

svels = [[1.1, 0],
         [-0.45, -0.45],
         [-0.4, 0.25]]

bodiess = [[Body(EARTH, EarthSize, EarthMass, [500, 350], [0, 0], main=True), Body(MOON, MoonSize, MoonMass, [500, 550], [-3.2, 0])],
           [Body(VENUS, VenusSize, VenusMass, [500, 350], [0, 0], main=True)],
           [Body(MARS, MarsSize, MarsMass, [500, 350], [0, 0], main=True), Body(PHOBOS, PhobosSize * 150, PhobosMass, [500, 397], [-2.5, 0]), Body(DEIMOS, DeimosSize * 150, DeimosMass, [500, 467], [-1.45, 0])]]

FONTFM = pygame.font.Font(convert('data/denmark.ttf'), 60)
FONTFT = pygame.font.Font(convert('data/denmark.ttf'), 110)
FONTFS = pygame.font.Font(convert('data/denmark.ttf'), 20)
FONTFI = pygame.font.Font(convert('data/denmark.ttf'), 20)
FONTFO = pygame.font.Font(convert('data/denmark.ttf'), 30)


def TextPassing(image):
    h = image.get_height()
    for i in range(WIDTH - 10, -image.get_width() - 10, -2):
        S.fill((0, 0, 0))
        S.blit(image, (i, 50))
        pygame.display.update((0, 50, WIDTH, h))
        Clock.tick(TDFPS)

def ShowScore(font, score):
    image = font.render('Score: ' + str(round(score)), 1, (127, 127, 127))
    S.blit(image, (Middle(image), 100))
    pygame.display.update()
    time.sleep(2)

TITLE = FONTFT.render('Stable Orbit', 1, (255, 255, 255))
MENU = ['Introduction', 'Play', 'Instructions', 'Exit']
MLINES = []
for i in range(0, len(MENU), 1):
    MLINES.append(FONTFM.render(MENU[i], 1, (90, 90, 90)))

MLINESS = []
for i in range(0, len(MENU), 1):
    MLINESS.append(FONTFM.render(MENU[i], 1, (180, 180, 180)))

sim = FONTFS.render('[SIMULATION]', 1, (200, 200, 200))
outta = FONTFO.render('"We\'re out of time, we\'ve got to launch now!"', 1, (200, 200, 200))
gl = FONTFO.render('"It works: let\'s launch!" Press SHIFT-ENTER.', 1, (200, 200, 200))
col = FONTFO.render('"Your moon collided with another body. Try adjusting its orbit."', 1, (200, 200, 200))
dri = FONTFO.render('"Your moon drifted out of orbit. Try reducing its velocity or moving it inwards."', 1, (200, 200, 200))
fal = FONTFO.render('"Your orbit was not stable. The human race is doomed to extinction. Better luck next time!"', 1, (200, 200, 200))
suc = FONTFO.render('"Your moon is safely in orbit... Years have passed and it\'s full as the population grows. You must build another."', 1, (200, 200, 200))
hyp = FONTFO.render('[Simulation speed increased]', 1, (200, 200, 200))
ffw = FONTFO.render('[Fast-forwarding time]', 1, (200, 200, 200))


transitions = [FONTFO.render('"We\'ve run out of space around Earth. The next moon is going to have to be built around Venus. Because of the distance there, We can\'t change the starting position of any launched moons."', 1, (200, 200, 200)),
               FONTFO.render('"Venus is full too. Destination: Mars. We can now adjust the starting position but at a cost of velocity. Mars has much weaker gravity. You\'ll need closer or slower orbits."', 1, (200, 200, 200)),
               FONTFO.render('"Well done. You have launched your moons quickly and accurately. Earth has recovered and people are slowly returning to it. Thanks to you, humanity lives on."', 1, (200, 200, 200))]



# Images from text
storyim = TextToPicture(story, (90, 90, 90), 25, convert('data/denmark.ttf'), 32)
instrim = TextToPicture(instructions, (90, 90, 90), 25, convert('data/denmark.ttf'), 32)

selop = 0
ext = False
while not ext:
    for event in pygame.event.get():
        if event.type == KEYDOWN:
            if event.key == K_DOWN:
                CLICK.play()
                selop += 1
                selop %= 4
            elif event.key == K_UP:
                CLICK.play()
                selop += 3
                selop %= 4
            elif event.key == K_ESCAPE:
                Stop()
            else:
                CLICK.play()
                if selop == 2:
                    si = False
                    while not si:
                        S.fill((0, 0, 0))
                        S.blit(instrim, (20, 30))
                        for event in pygame.event.get():
                            if event.type == KEYDOWN:
                                si = True
                            elif event.type == QUIT:
                                Stop()
                        pygame.display.update()
                    CLICK.play()
                elif selop == 3:
                    CLICK.play()
                    Stop()

                elif selop == 0:
                    si = False
                    while not si:
                        S.fill((0, 0, 0))
                        S.blit(storyim, (20, 30))
                        for event in pygame.event.get():
                            if event.type == KEYDOWN:
                                si = True
                            elif event.type == QUIT:
                                Stop()
                        pygame.display.update()
                    CLICK.play()
                elif selop == 1:
                    score = 0
                    qtm = False
                    zone = 0
                    while not qtm:
                        if zone > 2:
                            break
                        bodies = bodiess[zone][:]

                        for lv in range(0, min(len(levels), nl[zone]), 1):
                            if zone == 1:
                                bodies.append(Body(NEWMOON, levels[lv][0], levels[lv][1], sposs[lv], [0, 0]))
                            else:
                                bodies.append(Body(NEWMOON, levels[lv][0], levels[lv][1], [100, 100], [0, 0]))
                            if zone == 2:
                                speed = svels[lv]
                            else:
                                speed = [0, 0]

                            done = False
                            holding = False
                            m = [False] * 4
                            starttime = time.time()
                            while not done:
                                for event in pygame.event.get():
                                    if event.type == MOUSEBUTTONDOWN and zone != 1:
                                        if between((pygame.mouse.get_pos(), bodies[-1].position)) < (bodies[-1].size * (PSIZE / 2)):
                                            CLICK.play()
                                            holding = True
                                    elif event.type == MOUSEBUTTONUP:
                                        holding = False
                                    elif event.type == KEYDOWN:
                                        if event.key == K_UP:
                                            m[0] = True
                                        elif event.key == K_RIGHT:
                                            m[1] = True
                                        elif event.key == K_DOWN:
                                            m[2] = True
                                        elif event.key == K_LEFT:
                                            m[3] = True
                                        elif event.key == K_SPACE or event.key == K_RETURN:
                                            CLICK.play()
                                            S.fill((0, 0, 0))
                                            S.blit(LOADING, (Middle(LOADING), 200))
                                            pygame.display.update()
                                            Clock.tick(1)
                                            for b in bodies:
                                                b.BackUp()
                                            reali = False
                                            if pygame.key.get_pressed()[303] or pygame.key.get_pressed()[304]:
                                                reali = True
                                                d = Level(bodies, real=True)
                                            else:
                                                d = Level(bodies)
                                            if d:
                                                if d[0] == 'Success':
                                                    if reali:
                                                        bodies = d[1][:]
                                                        for b in bodies:
                                                            b.BackUp()
                                                        if lv < nl[zone] - 1:
                                                            TextPassing(suc)
                                                        score += max((0, lts[lv] + starttime - time.time()))
                                                        done = True
                                                        break
                                                    else:
                                                        TextPassing(gl)
                                                        time.sleep(1)

                                                elif d[0] == 1:
                                                    EXPLOSIONSOUND.play()
                                                    S.blit(EXPLOSION, (d[1][0] - 24, d[1][1] - 24))
                                                    pygame.display.update()
                                                    if reali:
                                                        TextPassing(fal)
                                                        score -= 1000
                                                        qtm = True
                                                    else:
                                                        TextPassing(col)

                                                elif d[0] == 2:
                                                    if reali:
                                                        TextPassing(fal)
                                                        score -= 1000
                                                        qtm = True
                                                    else:
                                                        TextPassing(dri)

                                            if not done:
                                                for b in bodies:
                                                    b.Restore()
                                            m = [False] * 4
                                        elif event.key == K_ESCAPE:
                                            qtm = True
                                    elif event.type == KEYUP:
                                        if event.key == K_UP:
                                            m[0] = False
                                        elif event.key == K_RIGHT:
                                            m[1] = False
                                        elif event.key == K_DOWN:
                                            m[2] = False
                                        elif event.key == K_LEFT:
                                            m[3] = False
                                    elif event.type == QUIT:
                                        Stop()
                                if done:
                                    break

                                if speed[0] > 8:
                                    m[1] = False
                                elif speed[0] < -8:
                                    m[3] = False

                                if speed[1] > 8:
                                    m[2] = False
                                elif speed[1] < -8:
                                    m[0] = False

                                if zone != 2:
                                    if m[0]:
                                        speed[1] -= 0.02
                                    if m[1]:
                                        speed[0] += 0.02
                                    if m[2]:
                                        speed[1] += 0.02
                                    if m[3]:
                                        speed[0] -= 0.02

                                bodies[-1].SetSpeed(speed[:])

                                if time.time() - lts[lv] > starttime:
                                    S.fill((0, 0, 0))
                                    TextPassing(outta)
                                    time.sleep(1)
                                    for b in bodies:
                                        b.BackUp()
                                    d = Level(bodies, real=True)
                                    if d:
                                        if d[0] == 'Success':
                                            bodies = d[1][:]
                                            for b in bodies:
                                                b.BackUp()
                                            done = True
                                            break
                                        else:
                                            score -= 1000
                                            qtm = True
                                    if not done:
                                        for b in bodies:
                                            b.Restore()
                                if qtm:
                                    break
                                    m = [False] * 4
                                if holding:
                                    bodies[-1].SetPos(pygame.mouse.get_pos())
                                S.blit(BACKGROUND, (-(BIGNESS // 2), -(BIGNESS // 2)))
                                for body in bodies:
                                    S.blit(body.texture, (body.position[0] - round(body.size * (PSIZE / 2)), body.position[1] - round(body.size * (PSIZE / 2))))
                                    S.blit(FONTFI.render('Time Remaining: ' + str(round(starttime + lts[lv] - time.time())), 1, (200, 200, 200)), (5, 5))
                                dsp = (round((speed[0] * 20) + bodies[-1].position[0]), round((speed[1] * 20) + bodies[-1].position[1]))
                                pygame.draw.circle(S, (255, 0, 0), dsp, 2)
                                pygame.draw.line(S, (255, 0, 0), dsp, bodies[-1].position, 1)
                                pygame.display.update()
                                Clock.tick(MFPS)
                        score += 1000
                        if not qtm:
                            TextPassing(transitions[zone])
                            zone += 1
                    ShowScore(FONTFM, score)
        elif event.type == QUIT:
            Stop()
    S.blit(BACKGROUND, (-(BIGNESS // 2), -(BIGNESS // 2)))
    S.blit(TITLE, (Middle(TITLE), 40))
    for i in range(0, len(MENU), 1):
        dloc = (100, 225 + (85 * i))
        if selop == i:
            S.blit(MLINESS[i], dloc)
        else:
            S.blit(MLINES[i], dloc)
    pygame.display.update()
